"""
Create the summary Excel file based on the individual Wormcat runs
"""
import pandas as pd


def create_category_summary(data, category_name):
    """
    Utility function create_category_summary
    """
    category = data[category_name].value_counts()
    category = pd.DataFrame(
        {category_name: category.index, 'Count': category.values})
    category = category.sort_values(by=[category_name])
    return category


def process_category_file_row(row, sheet):
    """
    Utility function process_category_file_row
    """
    file_name = row['file']
    label_category = f"Category {row['category']}"
    label_pvalue = f"{row['label']}_PValue"
    label_rgs = f"{row['label']}_RGS"

    cat_results = pd.read_csv(file_name)
    cat_results.rename(columns={'Category': label_category,
                       'RGS': label_rgs, 'PValue': label_pvalue}, inplace=True)
    cat_results.drop('Unnamed: 0', axis=1, inplace=True)
    cat_results.drop('AC', axis=1, inplace=True)

    sheet = pd.merge(sheet, cat_results, on=label_category, how='outer')
    sheet[label_pvalue] = sheet[label_pvalue].apply(significant)
    sheet[label_rgs] = sheet[label_rgs].apply(lambda x: x if x > 0 else 0)
    return sheet


def significant(value):
    """
    Utility function to return abreviations for significat values
    NV = Not a Value
    NS = Not Significant
    """
    ret_val = 'NV'
    if value < 0.05:
        ret_val = value
    elif value >= 0.05:
        ret_val = 'NS'
    return ret_val


def create_legend(writer, values, excel_formats):
    """
    Creates a simple sheet page as a Legend
    """
    data = {'Color Code <=': values[::-1]}
    legend_sheet = pd.DataFrame(data)

    legend_sheet.to_excel(writer, sheet_name='Legend', index=False)
    worksheet = writer.sheets['Legend']
    num_rows, num_columns = legend_sheet.shape
    sheet_range = f"A1:{chr(num_columns + 64)}{num_rows+1}"

    worksheet.conditional_format(sheet_range, {
                                 'type': 'cell', 'criteria': '=', 'value': 0, 'format': excel_formats[0]})
    for index, value in enumerate(values):
        worksheet.conditional_format(sheet_range, {
                                     'type': 'cell', 'criteria': '<=', 'value': value, 'format': excel_formats[index+1]})

    worksheet.autofit()


def process_category_files(files_to_process, annotation_file, out_data_xlsx):
    """
    Processes each category file and create the corasponding Excel summary sheet
    """
    data = pd.read_csv(annotation_file)
    #writer = pd.ExcelWriter("{}".format(out_data_xlsx), engine='openpyxl')
    writer = pd.ExcelWriter(f"{out_data_xlsx}", engine='xlsxwriter')

    sheets = files_to_process['sheet'].unique()
    values = [0.0000000001, 0.00000001, 0.000001, 0.0001, 0.05]
    formats = [{'bg_color': 'white', 'font_color': 'black', 'num_format': '0'},
               {'bg_color': '#244162', 'font_color': 'white',
                   'num_format': '0.000E+00'},
               {'bg_color': '#355f91', 'font_color': 'white',
                   'num_format': '0.000E+00'},
               {'bg_color': '#95b3d7', 'font_color': 'black',
                   'num_format': '0.000E+00'},
               {'bg_color': '#b8cce4', 'font_color': 'black',
                   'num_format': '0.000E+00'},
               {'bg_color': '#f4f2fe', 'font_color': 'black',
                   'num_format': '0.000E+00'}
               ]
    excel_formats = [writer.book.add_format(f) for f in formats]

    create_legend(writer, values, excel_formats)

    for sheet_label in sheets:
        cat_files = files_to_process[files_to_process['sheet'] == sheet_label]
        label_category = f"Category {cat_files['category'].iloc[0]}"
        category_sheet = create_category_summary(data, label_category)
        for index, row in cat_files.iterrows():
            category_sheet = process_category_file_row(row, category_sheet)

        category_sheet.to_excel(writer, sheet_name=sheet_label, index=False)
        worksheet = writer.sheets[sheet_label]
        num_rows, num_columns = category_sheet.shape
        sheet_range = f"B1:{chr(num_columns + 64)}{num_rows+1}"
        # print(f"{range=}")

        worksheet.conditional_format(sheet_range, {
                                     'type': 'cell', 'criteria': '=', 'value': 0, 'format': excel_formats[0]})
        for index, value in enumerate(values):
            worksheet.conditional_format(sheet_range, {
                                         'type': 'cell', 'criteria': '<=', 'value': value, 'format': excel_formats[index+1]})

        worksheet.autofit()

    writer.close()



"""
Create sunburt HTML file with Regulated Gene Set Data (RGS)
"""
import json
import os
import pkg_resources
import pandas as pd


def read_rgs_and_categories(file_nm_in):
    """
    Read the RGS Data and creat JSON based on Category 3 info
    """
    nodes_dict = {}
    try:
        df = pd.read_csv(file_nm_in)
        node1_list= []
        nodes_dict = {"name":"rgs", "children":node1_list}

        cat3 = df.groupby(["Category.3"]).count()

        cat3_dict={}
        for cat3_index, row in cat3.iterrows():
            count = int(row['Wormbase.ID'])
            cat3_dict[cat3_index] = count

        for key in cat3_dict:
            components = key.split(':')
            size = int(cat3_dict[key])
            #print("key {} size {}".format(key, size))
            if len(components) == 1:
                #print("add to node1 {}".format(components[0]))
                node1_list.append({"name": components[0].strip(), "size": size})
            elif len(components) == 2:
                node_list = getChildrenFor(components[0].strip(),nodes_dict)
                node_list.append({"name": components[1].strip(), "size": size})
            else:
                node_list = getChildrenFor2(components[0].strip(),components[1].strip(),nodes_dict)
                node_list.append({"name": components[2].strip(), "size": size})
    except Exception as e:
        print("Error in read_rgs_and_categories:", e)
        pass
    return nodes_dict


def create_sunburst(dir_nm):
    """
    1. Read the sunburst template file 
    2. Find the line "insert json here" in the template
    3. Inject the Cat3 JSON Data at this point in the file
    4. Save the file as sunburst.html
    """
    file_nm_sunburst_html = f"{dir_nm}{os.path.sep}sunburst.html"
    file_nm_rgs_and_categories = f"{dir_nm}{os.path.sep}rgs_and_categories.csv"
    data = read_sunburst_template()
    json_data = json.dumps(read_rgs_and_categories(file_nm_rgs_and_categories))
    

    with open(file_nm_sunburst_html, "w") as file:
        for d in data:
            file.write(d)
            if "insert json here" in d:
                json_data = json.dumps(read_rgs_and_categories(file_nm_rgs_and_categories))
                json_var = f"var json_data = {json_data}"
                file.write(json_var)



def getChildrenFor(parent, nodes_dict):
    """
    Utility function getChildrenFor given parent
    """
    children = nodes_dict['children']
    node_list = None
    for key in children:
        if parent == key['name']:
            if 'children' in key:
                node_list = key['children']
            break

    if node_list is None:
        node_list = []
        children.append({"name":parent, "children":node_list})
    return node_list


def getChildrenFor2(grand_parent, parent, nodes_dict):
    """
    Utility function getChildrenFor given grand parent
    """
    children = getChildrenFor(grand_parent, nodes_dict)
    node_list = None
    for key in children:
        if parent == key['name']:
            if 'children' in key:
                node_list = key['children']
            break

    if node_list is None:
        node_list = []
        children.append({"name":parent, "children":node_list})

    return node_list

def read_sunburst_template():
    """
    Utility function to fined the template file in the package directory
    """
    try:
        data = []
        #data = pkg_resources.resource_string(__name__, 'sunburst.template').decode('utf-8')
        data_file = pkg_resources.resource_filename(__name__, 'sunburst.template')
        with open(data_file, "r") as file:
            data = file.readlines()
        return data
    except Exception as e:
        print("Error reading sunburst_template:", e)
        return None


if __name__ == "__main__":
    print(f"starting {os.getcwd()}")
    create_sunburst('RGS_Feb-14-2020-11_45_54')





"""
Module provides functionality for running Wormcat in batch mode at the command linr
"""
import os
import sys
import argparse
import shutil
import zipfile
import importlib.metadata
from datetime import datetime
import warnings
import pandas as pd
from wormcat_batch.execute_r import ExecuteR
from wormcat_batch.create_wormcat_xlsx import process_category_files
from wormcat_batch.create_sunburst import create_sunburst

warnings.filterwarnings('ignore', category=UserWarning, module='openpyxl')
warnings.simplefilter(action='ignore', category=FutureWarning)


def extract_csv_files(input_excel_nm, csv_file_path):
    """
    Create CSV Files from the given Excel spreadsheet
    """
    try:
        input_excel = pd.ExcelFile(input_excel_nm)
    except ValueError:
        print(f"ERROR: File Name [{input_excel_nm}[] is Not a Valid Excel File.")
        sys.exit(-1)
    
    for sheet in input_excel.sheet_names:
        sheet_df = input_excel.parse(sheet)
        sheet_df.to_csv(
            f'{csv_file_path}{os.path.sep}{sheet}.csv', index=False)


def process_csv_files(csv_file_path, wormcat_out_path, annotation_file):
    """
    Read the csv data files and process each individually through Wormcat
    """
    for dir_content in os.listdir(csv_file_path):
        if dir_content.endswith(".csv"):
            content_full_path = os.path.join(csv_file_path, dir_content)
            if os.path.isfile(content_full_path):
                header_line = None
                first_line = None
                with open(content_full_path, 'r', encoding='utf-8') as file:
                    header_line = file.readline().strip()
                    first_line = file.readline().strip()

                # This is a hack to check for Wormbase IDs if we see a Wormbase.ID in the first line ignoring the header
                if 'WBGene' in first_line[:16]:
                    header_line = "Wormbase.ID"
                wormcat_input_type = header_line.replace(' ', '.')
                csv_file_nm = os.path.basename(content_full_path)
                file_nm_wo_ext = csv_file_nm[:-4]  # Remove .csv from file name
                title = file_nm_wo_ext.replace('_', ' ')
                wormcat_output_dir = f'{wormcat_out_path}{os.path.sep}{file_nm_wo_ext}'
                execute_r = ExecuteR()
                execute_r.worm_cat_fun(
                    content_full_path, wormcat_output_dir, title, annotation_file, wormcat_input_type)
                create_sunburst(wormcat_output_dir)
    return wormcat_out_path


def create_summary_spreadsheet(wormcat_out_path, annotation_file, out_xsl_file_nm):
    """
    After all the sheets on the Excel have been executed or CSV files processed 
    create a dataframe that can be used to summarize the results.
    This dataframe is used to create the output Excel.
    """
    process_lst = []
    for dir_nm in os.listdir(wormcat_out_path):
        if os.path.isdir(os.path.join(wormcat_out_path, dir_nm)):
            for cat_num in [1, 2, 3]:
                rgs_fisher = f"{wormcat_out_path}{os.path.sep}{dir_nm}{os.path.sep}rgs_fisher_cat{cat_num}.csv"
                cat_nm = f"Cat{cat_num}"
                row = {'sheet': cat_nm, 'category': cat_num,
                    'file': rgs_fisher, 'label': dir_nm}
                process_lst.append(row)

    df_process = pd.DataFrame(process_lst, columns=[
                              'sheet', 'category', 'file', 'label'])
    process_category_files(df_process, annotation_file, out_xsl_file_nm)

# Wormcat Utility functions


def get_wormcat_lib():
    """
    Find the location where the R Wormcat program is installed
    """
    execute_r = ExecuteR()
    path = execute_r.wormcat_library_path_fun()
    if path:
        first_quote = path.find('"')
        last_quote = path.rfind('"')
        if last_quote == -1:
            print("Wormcat is not installed or cannot be found.")
            exit(-1)
        path = path[first_quote+1:last_quote]
    # print(f"wormcat_lib_path={path}")
    return path


def get_category_files(path):
    """
    Get the list of available annotation files for Wormcat.
    These files exist in the R wormcat install under the "extdata" directory
    and end with .csv
    """
    category_files = []
    path = f"{path}{os.path.sep}extdata"
    for root, dirs, files in os.walk(path):
        for filename in files:
            if filename.endswith(".csv"):
                category_files.append(filename)

    return category_files

# Utility functions


def create_directory(directory, with_backup=False):
    """
    Utility function to create a directory and backup the original if it exists and has content
    And the content was put there by a previous run of the program
    """
    timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")
    created_by_flag = '.created_by_wormcat_batch'

    if with_backup and os.path.exists(directory) and os.listdir(directory):
        created_by_flag_check = os.path.join(directory, created_by_flag)
        if os.path.exists(created_by_flag_check):
            print(f"Creating backup of existing directory [{directory}]")
            # Create backup directory name with a unique timestamp suffix
            backup_dir = f"{directory}_{timestamp}.bk"
            shutil.move(directory, backup_dir)
        else:
            print("ERROR")
            print(f"Directory {directory} is not Empty and was not created by a prior run of this program.")
            print("Change the output directory to an empty or non exsisting directory.")
            sys.exit(-1)

    try:
        os.makedirs(directory, exist_ok=True)
        if with_backup:
            created_by_flag_check = os.path.join(directory, created_by_flag)
            with open(created_by_flag_check, 'w', encoding='utf-8') as file:
                file.write(timestamp)
    except OSError as err:
        print(f"ERROR: Cannot create directory\n{err}")
        sys.exit(-1)



def zip_directory(directory_path, zip_file_name):
    """
    Compress the content of a directory in zip format.
    """
    with zipfile.ZipFile(zip_file_name, 'w') as zipf:
        for root, _, files in os.walk(directory_path):
            for file in files:
                file_path = os.path.join(root, file)
                zipf.write(file_path, os.path.relpath(
                    file_path, directory_path))


def is_valid_directory_name(directory_name):
    """
    Check if the directory name string is a valid name
    """
    ret_val = True
    if not isinstance(directory_name, str):
        ret_val = False

    # Check if the directory name is empty or contains only whitespace
    if not directory_name.strip():
        ret_val = False

    # Check if the directory name contains any invalid characters
    invalid_characters = '\'?%*|"<>'
    if any(char in invalid_characters for char in directory_name):
        ret_val = False

    return ret_val

class WormcatArgumentParser(argparse.ArgumentParser):
    """
    Add annotation file descriptions to help text
    """
    def print_help(self, file=None):
        wormcat_path = get_wormcat_lib()
        annotation_files = get_category_files(wormcat_path)
        annotation_file_names = ""
        for index, annotation_file in enumerate(sorted(annotation_files)):
            annotation_file_names += f"\t{index+1} {annotation_file}\n"

        super().print_help(file)
        print(f"\nannotation-files:\n{annotation_file_names}")

def print_error_and_quit(parser, error_msg):
    """
    Print command line ERROR message and quit
    """
    parser.print_usage()
    print(f"ERROR: {error_msg}")
    sys.exit(-1)

def process_command_arguments():
    """
    The process_command_arguments method validates and sets the input arguments 
    to conform with downstream processing
    """
    parser = WormcatArgumentParser()
    parser.add_argument('-i', '--input-excel',
                        help='Input file in Excel/Wormcat format')
    parser.add_argument('-c', '--input-csv-path',
                        help='Input path to a collection of CSV files in Wormcat format')
    parser.add_argument('-o', '--output-path', help='Output path')
    parser.add_argument('-a', '--annotation-file', default='whole_genome_v2_nov-11-2021.csv',
                        help='Annotation file name or path default=whole_genome_v2_nov-11-2021.csv')
    parser.add_argument('-t', '--clean-temp', default='True',
                        help='Remove files created while processing default=True')

    parser.add_argument('-v', '--version', action='version',
                        version=f'%(prog)s v{importlib.metadata.version("wormcat_batch")}')
    args = parser.parse_args()

    if not args.input_excel and not args.input_csv_path:
        print_error_and_quit(parser, "An Excel Input file or a path to CSV files is required.")

    if args.input_excel and args.input_csv_path:
        print_error_and_quit(parser, "--input-excel [-i] and  --input-csv-path [-c] can not be used at the same time.")
    
    if args.input_excel and not os.path.isfile(args.input_excel):
        print_error_and_quit(parser, f""" An Excel Input file not found.
        [{args.input_excel}] is not a valid file name for wormcat batch.""")
      
    if args.input_csv_path and not os.path.isdir(args.input_csv_path):
        print_error_and_quit(parser, f""" A CSV Directory was not found.
        [{args.input_csv_path}] is not a valid directory name for wormcat batch.""")

    if not args.output_path or not is_valid_directory_name(args.output_path):
        print_error_and_quit(parser, f""" An Output path is required with a valid name.
        [{args.output_path}] is not valid directory path for wormcat batch.""")

    # if args.annotation_file is not a path to an annotation file
    # vaildate the input name and set the full path
    if not os.path.sep in args.annotation_file:
        wormcat_path = get_wormcat_lib()
        annotation_files = get_category_files(wormcat_path)
        
        if not args.annotation_file or not args.annotation_file in annotation_files:
            print_error_and_quit(parser, "Missing or incorrect annotation-file-nm. See --help for details.")
            
        args.annotation_file = f"{wormcat_path}{os.path.sep}extdata{os.path.sep}{args.annotation_file}"

    elif args.annotation_file and os.path.sep in args.annotation_file:
        if not os.path.isfile(args.annotation_file):
            print_error_and_quit(parser, f"""Local Annotation file not found.
            [{args.annotation_file}] is not a valid file name for wormcat batch.""")

    # Support TRUE or True as input to clean temp
    # otherwise set to False
    if args.clean_temp.lower().title() == 'True':
        args.clean_temp = True
    else:
        args.clean_temp = False

    print(f"\tInput Excel     = {args.input_excel}")
    print(f"\tInput CSV Path  = {args.input_csv_path}")
    print(f"\tOutput Path     = {args.output_path}")
    print(f"\tAnnotation File = {args.annotation_file}")
    print(f"\tClean Temp      = {args.clean_temp}")
    print("\n")


    return args


def main():
    """
    Main control execution of Wormcat Batch
    """
    print("Starting Wormcat Batch")

    # Validate the command line arguments
    args = process_command_arguments()

    # Create the output directory if it does not exsist.
    # If it does exist and has content create a backup.
    create_directory(args.output_path, with_backup=True)

    # If we are given an input_csv_path use it.
    if args.input_csv_path:
        csv_file_path = args.input_csv_path
    else:
        # Else we MUST have been given an input_excel
        # Create a directory and extract the Excel sheets as csv data
        csv_file_path = f"{args.output_path}{os.path.sep}csv_files"
        create_directory(csv_file_path)
        extract_csv_files(args.input_excel, csv_file_path)

    # Create a directory based on the Excel file name or CSV directory and a timestamp
    # Wormcat processing will write to this directory
    if args.input_excel:
        base_input_excel = os.path.basename(args.input_excel)
        input_data_nm = os.path.splitext(base_input_excel)[0]
    else:
        input_data_nm = os.path.basename(args.input_csv_path)

    timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")
    output_base_dir = f"{input_data_nm}_{timestamp}"
    wormcat_out_path = f"{args.output_path}{os.path.sep}{output_base_dir}"
    create_directory(wormcat_out_path)

    # Call Wormcat on each CSV file
    process_csv_files(csv_file_path, wormcat_out_path, args.annotation_file)

    # Create a summary spreadsheet of all CSV runs
    out_xsl_file_nm = f"{wormcat_out_path}{os.path.sep}Out_{input_data_nm}.xlsx"
    create_summary_spreadsheet(
        wormcat_out_path, args.annotation_file, out_xsl_file_nm)

    # Zip the results
    zip_dir_nm = f"{args.output_path}{os.path.sep}{output_base_dir}.zip"
    zip_directory(wormcat_out_path, zip_dir_nm)

    # If set Remove files created while processing
    if args.clean_temp:
        shutil.rmtree(wormcat_out_path)
        if args.input_excel:
            shutil.rmtree(csv_file_path)


if __name__ == '__main__':
    main()



"""
Module to delegate calls to R program execution
"""
from subprocess import Popen, PIPE
import sys
import os
import platform
import logging

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

class ExecuteR(object):
    """
    Class to mange R exection
    """
    wormcat_r = f'{os.path.dirname(__file__)}{os.path.sep}worm_cat.R'

    worm_cat_function = [wormcat_r,
                               '--file', 0,
                               '--title', 1,
                               '--out_dir', 2,
                               '--rm_dir', 3,
                               '--annotation_file',4,
                               '--input_type', 5,
                               '--zip_files', 6
                               ]

    is_wormcat_installed = f'{os.path.dirname(__file__)}{os.path.sep}is_wormcat_installed.R'
    wormcat_library_path = [is_wormcat_installed, '--no-save', 0, '--quiet', 1]

    if platform.system() == 'Windows':
        wormcat_library_path.insert(0,'rscript.exe')
        worm_cat_function.insert(0,'rscript.exe')


    def wormcat_library_path_fun(self):
        """
        Utility function to get the R library path to the Wormcat function
        """
        ret_val = self.run(self.wormcat_library_path,"")
        return ret_val

    def worm_cat_fun(self, file_name, out_dir, title, annotation_file, input_type):
        """
        Utility function to setup the call to Wormcat function
        """
        print(f"Calling Wormcat with {os.path.basename(file_name)[:-4]} data")
        ret_val = self.run(self.worm_cat_function, file_name, title, out_dir, "False", annotation_file, input_type, "False")
        return ret_val

    def run(self, arg_list, *args):
        """
        Execute the R script
        """
        try:
            processed_args = self.process_args(arg_list, *args)
            process = Popen(processed_args, stdout=PIPE)
            out, err = process.communicate()
            out = str(out, 'utf-8')
            if not out:
                out = None
            #sys.stderr.write("run: out={} err={}\n".format(out,err))
            return out
        except Exception as err:
            sys.stderr.write(f"ERROR: command line error {args}\n")
            sys.stderr.write(f"ERROR: %{err}\n")
            sys.exit(-1)

    def process_args(self, arg_tuple, *args):
        """
        Utility function to build the arg list
        """
        arg_list = list(arg_tuple)
        for index in range(0, len(arg_list)):
            if type(arg_list[index]) == int:
                # substitue for args passed in
                if arg_list[index] < len(args):
                    arg_list[index] = args[arg_list[index]]
                # if we have more substitutions than args passed delete the extras
                else:
                    del arg_list[index - 1:]
                    break
        return arg_list



"""
Setup for pypi releases of wormcat_batch
"""
from setuptools import setup
from pathlib import Path

# rm -rf dist
# python setup.py sdist
# pip install dist/wormcat_batch-1.0.1.tar.gz
# twine check dist/*
# twine upload --repository pypi dist/*


this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name='wormcat_batch',
      version='1.1.10',
      description='Batch processing for Wormcat data',
      long_description_content_type="text/markdown",
      long_description=long_description,

      url='https://github.com/dphiggs01/Wormcat_batch',
      author='Dan Higgins',
      author_email='daniel.higgins@yahoo.com',
      license='MIT',

      packages=['wormcat_batch'],
      install_requires=['pandas','xlrd','openpyxl','xlsxwriter'],
      entry_points={
          'console_scripts': ['wormcat_cli=wormcat_batch.run_wormcat_batch:main'],
      },
      include_package_data=True,
      zip_safe=False)
